// Brian's PSF Estimator

// March 26, 1998.  Brian Jones bbjones@egret6.stanford.edu
// Version 1.1

// This routine will give you the Kalman filter variance of the estimated
// incident angle of an electon, when you supply the following arguments.
// Note that the PSF of the gamma ray should be a combination of the 
// covariances from the two electrons.  If f is the fraction of energy
// in electron 1, then

// \sigma_\gamma = f \sigma_1^2 + (1-f) \sigma_2^2
// under the "everything is Gaussian" approximation.


// energy:  Incident gamma energy in MeV
// pitch: The strip pitch of the SSDs in mm.
// gap: The gap between planes in mm.
// conv:  the converter thickness, in radiation lengths.
// ssdthick: the SSD thickness, *in radiation lenghts*.
// incang: the incident electron angle, in radians.
// numplanes: the number of planes that the electrons pass through.  In this 
//            version, both electrons are assumed to pass through the same number
//            of planes.


// References for Kalman Filtering:
// "A Physicist's Guide to Kalman Filters," Brian Jones, 
//  WWW:  http://egret0.stanford.edu/bbjones/expkal.pdf
// Brian Jones, PhD Thesis, coming soon to a library near you.


// standard includes
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <string.h>
#include <vector>

// Classes to be used:
#include "CLHEP/Matrix/Matrix.h"
#include "CLHEP/Matrix/SymMatrix.h"
#include "CLHEP/Matrix/DiagMatrix.h"
#include "CLHEP/Matrix/Vector.h"


// Function prototypes

inline double ms(double E, double X);
// Vector version of getvar:
double getvar(double energy, double pitch, vector<double> gap, 
	      vector<double> conv, double ssdthick, double incang, 
	      int numplanes);
// Partial vector version:
double getvar(double energy, double pitch, vector<double> gap, 
	      double conv, double ssdthick, double incang, 
	      int numplanes);
// Scalar version of getvar:
double getvar(double energy, double pitch, double gap,
	      double conv, double ssdthick, double incang,
	      int numplanes);




double getvar(double energy, double pitch, double gap,
	      double conv, double ssdthick, double incang,
	      int numplanes) {
  
  //  cout << "Constructing vectors..." << endl;

  vector<double> vconv;
  vector<double> vgap;

  for(int i=0; i<numplanes; i++) {
    vconv.push_back(conv);
    vgap.push_back(gap);
  }
  
  vconv[numplanes-1]=vconv[numplanes-2]=0;

  //  cout << "vconv: ";
//   for(int i=0; i<numplanes; i++)
//     cout << vconv[i] << " ";
//   cout << endl << "vgap: ";
//   for(int i=0; i<numplanes; i++)
//     cout << vgap[i] << " ";
//   cout << endl;

  return(getvar(energy, pitch, vgap, vconv, ssdthick, incang, numplanes));

}

//Partial Vector version
double getvar(double energy, double pitch, vector<double> gap,
	      double conv, double ssdthick, double incang,
	      int numplanes) {
  
  //  cout << "Constructing vectors..." << endl;

  vector<double> vconv;

  for(int i=0; i<numplanes; i++) {
    vconv.push_back(conv);
  }
  
  vconv[numplanes-1]=vconv[numplanes-2]=0;

  return(getvar(energy, pitch, gap, vconv, ssdthick, incang, numplanes));

}



double getvar(double energy, double pitch, vector<double> gap, 
	      vector<double> conv, double ssdthick, double incang, 
	      int numplanes) {

HepMatrix projC(2,2,0);
HepMatrix C(2,2,0);

HepMatrix Q(2,2,0);  // multiple scattering
HepMatrix G(2,2,0);  // inverse of measurement error
// Note that I will store H*G*H.T() in G, since that is the only parameter
// of interest for these calculations.  In practice, that means only the
// (1,1) component will have anything in it.


HepMatrix F(2,2,1); // Propagator

double theta; // multiple scattering sigma.

 const double inf=1e10;
 const double Siradlen=50.0; // mm.  Very approximate.
 const double Pbradlen=0.0254; // mm. Way approximate.

int bad;  // to check for bad inverses.


// Stuff that isn't going to change at each plane
ssdthick=ssdthick/cos(incang); // first order incident angle corrections
 for(int i=0; i<conv.size(); i++) {
   conv[i]=conv[i]/cos(incang);
   gap[i]=gap[i]/cos(incang);
 }
pitch=pitch*cos(incang);
G(1,1)=12/(pitch*pitch);  // Inverse of pitch/root(12), squared.

if(numplanes<3)
  return(-1); // do not allow tracks with 2 or fewer hits

// We'll start at the bottom, so we can get the top variance in one pass.

// cout << endl << "Number of planes: " << numplanes << endl;
// Initial C matrix is infinite.
C(1,1)=inf;
C(2,2)=inf;
// Make first measurement
projC=C.inverse(bad) + G;
C=projC.inverse(bad);

//  cout << "Initial C matrix: " << C; 

for(int i=numplanes-1; i>0; i--) {
  F(1,2)=gap[i];
  theta=ms(energy, ssdthick+conv[i]);
  Q(1,1)=(ssdthick*Siradlen + conv[i]*Pbradlen)*theta; 
  Q(1,1)=Q(1,1)*Q(1,1)/3.;  // y displacement
  
  Q(2,1)=(ssdthick*Siradlen+ conv[i]*Pbradlen)*theta*theta/2.;
  Q(1,2)=Q(2,1);
  
  Q(2,2)=theta*theta;
  
  //  cout << "Plane " << i << " projected from Plane " << i+1 << endl;
  //  cout << "Gap between Planes " << i << " and " << i+1 << ": " << gap[i] << endl;
  projC=F * C * F.T() + Q;
  //  cout << "projC" << endl;
  //  cout << projC;
  
  //  cout << "G:" << G;
  C=projC.inverse(bad) + G;
  if(bad)
    cout << "projC has no inverse!" << projC;
  //  cout << "Inverse C: " << C;
  C=C.inverse(bad);
  if(bad)
    cout << "C has no inverse!" << C;
  
  //  cout << "C" << C;
}

//cout << endl << endl << "Covariance Matrix: " << C;
return(C(2,2));
}

inline double ms(double E, double X) {
  double bob=(13.6/E)*sqrt(X)*(1. + 0.038 * log(X) );
return(bob);
}
